#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.pilot_manager
   :platform: Unix
   :synopsis: Provides the interface for the PilotManager class.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import json
import urllib2

from radical.pilot.states import *
from radical.pilot.exceptions import *

from radical.pilot.object import Object
from radical.pilot.controller import PilotManagerController
from radical.pilot.compute_pilot import ComputePilot
from radical.pilot.utils.logger import logger
from radical.pilot.exceptions import * 

# -----------------------------------------------------------------------------
#
class PilotManager(Object):
    """A PilotManager holds :class:`radical.pilot.ComputePilot` instances that are
    submitted via the :func:`radical.pilot.PilotManager.submit_pilots` method.

    It is possible to attach one or more :ref:`chapter_machconf`
    to a PilotManager to outsource machine specific configuration
    parameters to an external configuration file.

    Each PilotManager has a unique identifier :data:`radical.pilot.PilotManager.uid`
    that can be used to re-connect to previoulsy created PilotManager in a
    given :class:`radical.pilot.Session`.

    **Example**::

        s = radical.pilot.Session(database_url=dbURL)

        pm1 = radical.pilot.PilotManager(session=s, resource_configurations=RESCONF)
        # Re-connect via the 'get()' method.
        pm2 = radical.pilot.PilotManager.get(session=s, pilot_manager_id=pm1.uid)

        # pm1 and pm2 are pointing to the same PilotManager
        assert pm1.uid == pm2.uid
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self, session, resource_configurations=None):
        """Creates a new PilotManager and attaches is to the session.

        .. note:: The `resource_configurations` (see :ref:`chapter_machconf`)
                  parameter is currently mandatory for creating a new
                  PilotManager instance.

        **Arguments:**

            * **session** [:class:`radical.pilot.Session`]:
              The session instance to use.

            * **resource_configurations** [`string` or `list of strings`]:
              A list of URLs pointing to :ref:`chapter_machconf`. Currently
              `file://`, `http://` and `https://` URLs are supported.

              If one or more resource_configurations are provided, Pilots
              submitted  via this PilotManager can access the configuration
              entries in the  files via the :class:`ComputePilotDescription`.
              For example::

                  pm = radical.pilot.PilotManager(session=s, resource_configurations="https://raw.github.com/radical-cybertools/radical.pilot/master/configs/futuregrid.json")

                  pd = radical.pilot.ComputePilotDescription()
                  pd.resource = "futuregrid.INDIA"  # defined in futuregrid.json
                  pd.cores    = 16
                  pd.runtime  = 5 # minutes

                  pilot = pm.submit_pilots(pd)

        **Returns:**

            * A new `PilotManager` object [:class:`radical.pilot.PilotManager`].

        **Raises:**
            * :class:`radical.pilot.PilotException`
        """
        self._session = session
        self._worker = None
        self._uid = None

        if resource_configurations == "~=RECON=~":
            # When we get the "~=RECON=~" keyword as resource_configurations,
            # we were called  from the 'get()' class method. In this case
            # object instantiation happens there...
            return

        ###############################
        # Create a new pilot manager. #
        ###############################

        # Donwload and parse the configuration file(s) and the content to
        # our resource dictionary.
        self._resource_cfgs = {}

        # Add 'localhost' as a built-in resource configuration
        self._resource_cfgs["localhost"] = {
            "URL":              "fork://localhost",
            "filesystem":       "file://localhost",
            "pre_bootstrap":    ["hostname", "date"],
            "task_launch_mode": "LOCAL",
        }

        if resource_configurations is not None:

            # implicit list conversion
            if not isinstance(resource_configurations, list):
                resource_configurations = [resource_configurations]

            for rcf in resource_configurations:
                try:
                    # download resource configuration file
                    response = urllib2.urlopen(rcf)
                    rcf_content = response.read()
                except urllib2.URLError, err:
                    msg = "Couln't open/download resource configuration file '%s': %s." % (rcf, str(err))
                    raise BadParameter(msg=msg)

                try:
                    # convert JSON string to dictionary and append
                    rcf_dict = json.loads(rcf_content)
                    for key, val in rcf_dict.iteritems():
                        if key in self._resource_cfgs:
                            raise BadParameter("Resource configuration entry for '%s' defined in %s is already defined." % (key, rcf))
                        self._resource_cfgs[key] = val
                except ValueError, err:
                    raise BadParameter("Couldn't parse resource configuration file '%s': %s." % (rcf, str(err)))

        # Start a worker process fo this PilotManager instance. The worker
        # process encapsulates database access, persitency et al.
        self._worker = PilotManagerController(
            pilot_manager_uid=None,
            pilot_manager_data={},
            db_connection=session._dbs)
        self._worker.start()

        self._uid = self._worker.pilot_manager_uid

        # Each pilot manager has a worker thread associated with it. The task
        # of the worker thread is to check and update the state of pilots, fire
        # callbacks and so on.
        self._session._pilot_manager_objects.append(self)
        self._session._process_registry.register(self._uid, self._worker)

    #--------------------------------------------------------------------------
    #
    def close(self, terminate=False):
        """Shuts down the PilotManager and its background workers in a 
        coordinated fashion.

        **Arguments:**

            * **terminate** [`bool`]: If set to True, all active pilots will 
              get canceled (default: False).

        """
        # Spit out a warning in case the object was already closed.
        if not self._uid:
            logger.warning("PilotManager object already closed.")
            return

        # If terminate is set, we cancel all pilots. 
        if terminate is True:
            self.cancel_pilots()

        # Shut down all worker processes if still active. stop() returns
        # only after a successful join(). 
        if self._worker is not None:
            # Stop the worker process
            self._worker.stop()
            # Remove worker from registry
            self._session._process_registry.remove(self._uid)

        logger.info("Closed PilotManager %s." % str(self._uid))
        self._uid = None

    #--------------------------------------------------------------------------
    #
    @classmethod
    def _reconnect(cls, session, pilot_manager_id):
        """PRIVATE: reconnect to an existing pilot manager.
        """
        uid_exists = PilotManagerController.uid_exists(
            db_connection=session._dbs,
            pilot_manager_uid=pilot_manager_id
        )

        if not uid_exists:
            raise BadParameter(
                "PilotManager with id '%s' not in database." % pilot_manager_id)

        obj = cls(session=session, resource_configurations="~=RECON=~")
        obj._uid = pilot_manager_id
        obj._resource_cfgs = None  # TODO: reconnect

        # Retrieve or start a worker process fo this PilotManager instance.
        worker = session._process_registry.retrieve(pilot_manager_id)
        if worker is not None:
            obj._worker = worker
        else:
            obj._worker = PilotManagerController(
                pilot_manager_uid=pilot_manager_id,
                pilot_manager_data={},
                db_connection=session._dbs)
            session._process_registry.register(pilot_manager_id, obj._worker)

        # start the worker if it's not already running
        if obj._worker.is_alive() is False:
            obj._worker.start()

        return obj

    # -------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object.
        """
        self._assert_obj_is_valid()

        object_dict = {
            'uid': self.uid
        }
        return object_dict

    # -------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the object.
        """
        return str(self.as_dict())

    # -------------------------------------------------------------------------
    #
    def submit_pilots(self, pilot_descriptions):
        """Submits a new :class:`radical.pilot.ComputePilot` to a resource.

        **Returns:**

            * One or more :class:`radical.pilot.ComputePilot` instances
              [`list of :class:`radical.pilot.ComputePilot`].

        **Raises:**

            * :class:`radical.pilot.radical.pilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Implicit list conversion.
        if not isinstance(pilot_descriptions, list):
            pilot_descriptions = [pilot_descriptions]

        # Itereate over the pilot descriptions, try to create a pilot for
        # each one and append it to 'pilot_obj_list'.
        pilot_obj_list = list()

        for pilot_description in pilot_descriptions:

            if pilot_description.resource is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'resource'."
                raise BadParameter(error_msg)

            elif pilot_description.runtime is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'runtime'."
                raise BadParameter(error_msg)

            elif pilot_description.cores is None:
                error_msg = "ComputePilotDescription does not define mandatory attribute 'cores'."
                raise BadParameter(error_msg)

            # Make sure resource key is known.
            if pilot_description.resource not in self._resource_cfgs:
                error_msg = "ComputePilotDescription.resource key '%s' is not known by this PilotManager." % pilot_description.resource
                raise BadParameter(error_msg)
            else:
                resource_cfg = self._resource_cfgs[pilot_description.resource]

            # If 'default_sandbox' is defined, set it.
            if pilot_description.sandbox is not None:
                if "valid_roots" in resource_cfg:
                    is_valid = False
                    for vr in resource_cfg["valid_roots"]:
                        if pilot_description.sandbox.startswith(vr):
                            is_valid = True
                    if is_valid is False:
                        raise BadParameter("Working directory for resource '%s' defined as '%s' but needs to be rooted in %s " % (pilot_description.resource, pilot_description.sandbox, resource_cfg["valid_roots"]))

            # After the sanity checks have passed, we can register a pilot
            # startup request with the worker process and create a facade
            # object.

            pilot = ComputePilot._create(
                pilot_description=pilot_description,
                pilot_manager_obj=self)

            pilot_uid = self._worker.register_start_pilot_request(
                pilot=pilot,
                resource_config=resource_cfg,
                session=self._session)

            pilot._uid = pilot_uid

            pilot_obj_list.append(pilot)

        # Implicit return value conversion
        if len(pilot_obj_list) == 1:
            return pilot_obj_list[0]
        else:
            return pilot_obj_list

    # -------------------------------------------------------------------------
    #
    def list_pilots(self):
        """Lists the unique identifiers of all :class:`radical.pilot.ComputePilot`
        instances associated with this PilotManager

        **Returns:**

            * A list of :class:`radical.pilot.ComputePilot` uids [`string`].

        **Raises:**

            * :class:`radical.pilot.radical.pilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Get the pilot list from the worker
        return self._worker.list_pilots()

    # -------------------------------------------------------------------------
    #
    def get_pilots(self, pilot_ids=None):
        """Returns one or more :class:`radical.pilot.ComputePilot` instances.

        **Arguments:**

            * **pilot_uids** [`list of strings`]: If pilot_uids is set,
              only the Pilots with  the specified uids are returned. If
              pilot_uids is `None`, all Pilots are returned.

        **Returns:**

            * A list of :class:`radical.pilot.ComputePilot` objects
              [`list of :class:`radical.pilot.ComputePilot`].

        **Raises:**

            * :class:`radical.pilot.radical.pilotException`
        """
        self._assert_obj_is_valid()

        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            pilot_ids = [pilot_ids]

        pilots = ComputePilot._get(pilot_ids=pilot_ids, pilot_manager_obj=self)
        return pilots

    # -------------------------------------------------------------------------
    #
    def wait_pilots(self, pilot_ids=None,
                    state=[DONE, FAILED, CANCELED],
                    timeout=None):
        """Returns when one or more :class:`radical.pilot.ComputePilots` reach a
        specific state or when an optional timeout is reached.

        If `pilot_uids` is `None`, `wait_pilots` returns when **all** Pilots
        reach the state defined in `state`.

        **Arguments:**

            * **pilot_uids** [`string` or `list of strings`]
              If pilot_uids is set, only the Pilots with the specified uids are
              considered. If pilot_uids is `None` (default), all Pilots are
              considered.

            * **state** [`list of strings`]
              The state(s) that Pilots have to reach in order for the call
              to return.

              By default `wait_pilots` waits for the Pilots to reach
              a **terminal** state, which can be one of the following:

              * :data:`radical.pilot.DONE`
              * :data:`radical.pilot.FAILED`
              * :data:`radical.pilot.CANCELED`

            * **timeout** [`float`]
              Optional timeout in seconds before the call returns regardless
              whether the Pilots have reached the desired state or not.
              The default value **-1.0** never times out.

        **Raises:**

            * :class:`radical.pilot.radical.pilotException`
        """
        self._assert_obj_is_valid()

        if not isinstance(state, list):
            state = [state]

        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            pilot_ids = [pilot_ids]

        start_wait = time.time()
        all_done = False
        return_states = []

        while all_done is False:

            all_done = True

            pilots_json = self._worker.get_compute_pilot_data()

            for pilot in pilots_json:
                if pilot['state'] not in state:
                    all_done = False
                    break  # leave for loop
                else:
                    return_states.append(pilot['state'])

            # check timeout
            if (None != timeout) and (timeout <= (time.time() - start_wait)):
                break

            # wait a bit
            time.sleep(1)

        # done waiting
        return return_states

    # -------------------------------------------------------------------------
    #
    def cancel_pilots(self, pilot_ids=None):
        """Cancels one or more ComputePilots.

        **Arguments:**

            * **pilot_uids** [`string` or `list of strings`]
              If pilot_uids is set, only the Pilots with the specified uids are
              canceled. If pilot_uids is `None`, all Pilots are canceled.

        **Raises:**

            * :class:`radical.pilot.radical.pilotException`
        """
        # Check if the object instance is still valid.
        self._assert_obj_is_valid()

        # Implicit list conversion.
        if (not isinstance(pilot_ids, list)) and (pilot_ids is not None):
            pilot_ids = [pilot_ids]

        # Register the cancelation request with the worker.
        self._worker.register_cancel_pilots_request(pilot_ids=pilot_ids)

    # -------------------------------------------------------------------------
    #
    def register_callback(self, callback_function):
        """Registers a new callback function with the PilotManager.
        Manager-level callbacks get called if any of the ComputePilots managed
        by the PilotManager change their state.

        All callback functions need to have the same signature::

            def callback_func(obj, state)

        where ``object`` is a handle to the object that triggered the callback
        and ``state`` is the new state of that object.
        """
        self._assert_obj_is_valid()

        self._worker.register_manager_callback(callback_function)



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler.RoundRobinScheduler
   :platform: Unix
   :synopsis: A multi-pilot, round-robin scheduler.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os 

from radical.pilot.utils.logger        import logger
from radical.pilot.scheduler.interface import Scheduler 

# -----------------------------------------------------------------------------
# 
class RoundRobinScheduler(Scheduler):
    """RoundRobinScheduler implements a multi-pilot, round-robin 
    scheduling algorithm.
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self):
        """Le constructeur.
        """
        Scheduler.__init__(self)
        logger.info("Loaded scheduler: %s." % self.name)

        self._idx = 0

    # -------------------------------------------------------------------------
    #
    def __del__(self):
        """Le destructeur.
        """
        if os.getenv("RADICALPILOT_GCDEBUG", None) is not None:
            logger.debug("__del__(): %s." % self.name)

    # -------------------------------------------------------------------------
    #
    def _name(self):
        return "RoundRobinScheduler"

    # -------------------------------------------------------------------------
    #
    def schedule(self, manager, unit_descriptions):
        # the scheduler will return a dictionary of the form:
        #   { 
        #     pilot_id_1  : [ud_1, ud_2, ...], 
        #     pilot_id_2  : [ud_3, ud_4, ...], 
        #     ...
        #   }
        # The scheduler may not be able to schedule some units -- those will
        # simply not be listed for any pilot.  The UM needs to make sure
        # that no UD from the original list is left untreated, eventually.

        #print "round-robin scheduling of %s units" % len(unit_descriptions)

        if  not manager :
            raise RuntimeError ('Unit scheduler is not initialized')

        pilots = manager.list_pilots ()
        ret    = dict()

        if not len (pilots) :
            raise RuntimeError ('Unit scheduler cannot operate on empty pilot set')


        for ud in unit_descriptions :
            
            if  self._idx >= len(pilots) : 
                self._idx = 0
            
            pilot = pilots[self._idx]

            if  pilot not in ret :
                ret[pilot] = []

            ret[pilot].append (ud)

            self._idx += 1


        return ret



#pylint: disable=C0301, C0103, W0212, E1101, R0903

"""
.. module:: radical.pilot.utils.version
   :platform: Unix
   :synopsis: Implementation of the version extractor.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os

version = open(os.path.dirname (os.path.abspath (__file__)) + "/../VERSION", 'r').read().strip()



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.context
   :platform: Unix
   :synopsis: Implementation of the Context class(es).

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import saga.context as sc

# ------------------------------------------------------------------------------
#
class SSHCredential(object):
    """An SSHCredential object represents an SSH identity.
    """ 

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :
        """Creates a new SSHCredential object.
        """
        self._context = sc.Context("SSH")

    #---------------------------------------------------------------------------
    #
    @classmethod
    def from_dict(cls, thedict):
        """Creates a new object instance from a string.

                c._from_dict(x.as_dict) == x
        """
        obj = cls()
        obj._context.user_id   = thedict["user_id"]
        obj._context.user_pass = thedict["user_pass"]
        obj._context.user_key  = thedict["user_key"]

        return obj

    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object.
        """
        dct = {
            "type"      : "SSH",
            "user_id"   : self._context.user_id,
            "user_pass" : self._context.user_pass,
            "user_key"  : self._context.user_key,
        }
        return dct

    # --------------------------------------------------------------------------
    #
    def __str__ (self):
        """Returns the string representation of the object.
        """
        return str(self.as_dict())

    # --------------------------------------------------------------------------
    #
    @property
    def user_id(self):
        """ TODO: Document me 
        """
        return self._context.user_id
    @user_id.setter
    def user_id(self, value):
        """ TODO: document me
        """
        self._context.user_id = value
    
    # --------------------------------------------------------------------------
    #
    @property
    def user_pass(self):
        """ TODO: Document me 
        """
        return self._context.user_pass
    @user_pass.setter
    def user_pass(self, value):
        """ TODO: Document me 
        """
        self._context.user_pass = value

    # --------------------------------------------------------------------------
    #
    @property
    def user_key(self):
        """ TODO: Document me 
        """
        return self._context.user_key
    @user_key.setter
    def user_key(self, value):
        """ TODO: Document me 
        """
        self._context.user_key = value



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.controller.unit_manager_controller
   :platform: Unix
   :synopsis: Implements a multiprocessing worker backend for
              the UnitManager class.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import Queue
import weakref
import threading

from multiprocessing import Pool

from radical.pilot.states import *

from radical.utils import which
from radical.pilot.utils.logger import logger

from radical.pilot.controller.input_file_transfer_worker import InputFileTransferWorker
from radical.pilot.controller.output_file_transfer_worker import OutputFileTransferWorker


# ----------------------------------------------------------------------------
#
class UnitManagerController(threading.Thread):
    """UnitManagerController handles backend interaction for the UnitManager 
    class. It is threaded and manages background worker processes. 
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, unit_manager_uid, scheduler, input_transfer_workers,
        output_transfer_workers, db_connection, db_connection_info):

        # Multithreading stuff
        threading.Thread.__init__(self)
        self.daemon = True

        # Stop event can be set to terminate the main loop
        self._stop = threading.Event()
        self._stop.clear()

        # Initialized is set, once the run loop has pulled status
        # at least once. Other functions use it as a guard.
        self._initialized = threading.Event()
        self._initialized.clear()

        # The shard_data_manager handles data exchange between the worker
        # process and the API objects. The communication is unidirectional:
        # workers WRITE to _shared_data and API methods READ from _shared_data.
        # The strucuture of _shared_data is as follows:
        #
        # { unit1_uid: MongoDB document (dict),
        #   unit2_uid: MongoDB document (dict),
        #   ...
        # }
        #
        self._shared_data = dict()
        self._shared_data_lock = threading.Lock()

        # The manager-level list.
        #
        self._manager_callbacks = list()

        # The MongoDB database handle.
        self._db = db_connection

        if unit_manager_uid is None:
            # Try to register the UnitManager with the database.
            self._um_id = self._db.insert_unit_manager(
                scheduler=scheduler,
                input_transfer_workers=input_transfer_workers,
                output_transfer_workers=output_transfer_workers)
            self._num_input_transfer_workers = input_transfer_workers
            self._num_output_transfer_workers = output_transfer_workers
        else:
            um_json = self._db.get_unit_manager(unit_manager_id=unit_manager_id)
            self._um_id = unit_manager_uid
            self._num_input_transfer_workers = um_json["input_transfer_workers"]
            self._num_output_transfer_workers = um_json["output_transfer_workers"]

        # The INPUT transfer worker(s) are autonomous processes that
        # execute input file transfer requests concurrently.
        self._input_file_transfer_worker_pool = []
        for x in range(1, self._num_input_transfer_workers+1):
            worker = InputFileTransferWorker(
                db_connection_info=db_connection_info, 
                unit_manager_id=self._um_id,
                number=x
            )
            self._input_file_transfer_worker_pool.append(worker)
            worker.start()

        # The OUTPUT transfer worker(s) are autonomous processes that
        # execute output file transfer requests concurrently.
        self._output_file_transfer_worker_pool = []
        for x in range(1, self._num_output_transfer_workers+1):
            worker = OutputFileTransferWorker(
                db_connection_info=db_connection_info, 
                unit_manager_id=self._um_id,
                number=x
            )
            self._output_file_transfer_worker_pool.append(worker)
            worker.start()

    # ------------------------------------------------------------------------
    #
    @classmethod
    def uid_exists(cls, db_connection, unit_manager_uid):
        """Checks wether a particular unit manager UID exists.
        """
        exists = False

        if unit_manager_uid in db_connection.list_unit_manager_uids():
            exists = True

        return exists

    # ------------------------------------------------------------------------
    #
    @property
    def unit_manager_uid(self):
        """Returns the uid of the associated UnitManager
        """
        return self._um_id

    # ------------------------------------------------------------------------
    #
    def stop(self):
        """stop() signals the process to finish up and terminate.
        """
        self._stop.set()
        self.join()
        logger.debug("Worker thread (ID: %s[%s]) for UnitManager %s stopped." %
                    (self.name, self.ident, self._um_id))

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_data(self, unit_uid):
        """Retruns the raw data (json dicts) of one or more ComputeUnits
           registered with this Worker / UnitManager
        """
        # Wait for the initialized event to assert proper operation.
        self._initialized.wait()

        return self._shared_data[unit_uid]["data"]

    # ------------------------------------------------------------------------
    #
    def call_callbacks(self, unit_id, new_state):
        """Wrapper function to call all all relevant callbacks, on unit-level
        as well as manager-level.
        """
        for cb in self._shared_data[unit_id]['callbacks']:
            try:
                cb(self._shared_data[unit_id]['facade_object'],
                   new_state)
            except Exception, ex:
                logger.error(
                    "Couldn't call callback function %s" % str(ex))

        # If we have any manager-level callbacks registered, we
        # call those as well!
        for cb in self._manager_callbacks:
            try:
                cb(self._shared_data[unit_id]['facade_object'],
                   new_state)
            except Exception, ex:
                logger.error(
                    "Couldn't call callback function %s" % str(ex))

    # ------------------------------------------------------------------------
    #
    def _set_state(self, unit_uid, state, log):

        if not isinstance(log, list):
            log = [log]

        # Acquire the shared data lock.
        self._shared_data_lock.acquire()

        old_state = self._shared_data[unit_uid]["data"]["state"]

        # Update the database.
        self._db.set_compute_unit_state(unit_uid, state, log)

        # Update shared data.
        self._shared_data[unit_uid]["data"]["state"] = state
        self._shared_data[unit_uid]["data"]["statehistory"] = state
        self._shared_data[unit_uid]["data"]["log"].extend(log)

        # Call the callbacks
        if state != old_state:
            # On a state change, we fire zee callbacks.
            logger.info(
                "ComputeUnit '%s' state changed from '%s' to '%s'." %
                (unit_uid, old_state, state)
            )

            # The state of the unit has changed, We call all
            # unit-level callbacks to propagate this.
            self.call_callbacks(unit_uid, state)

        # Release the shared data lock.
        self._shared_data_lock.release()

    # ------------------------------------------------------------------------
    #
    def run(self):
        """run() is called when the process is started via
           PilotManagerController.start().
        """
        logger.debug("Worker thread (ID: %s[%s]) for UnitManager %s started." %
                    (self.name, self.ident, self._um_id))

        # transfer results contains the futures to the results of the
        # asynchronous transfer operations.
        transfer_results = list()

        while not self._stop.is_set():

            # =================================================================
            #
            # Check and update units. This needs to be optimized at
            # some point, i.e., state pulling should be conditional
            # or triggered by a tailable MongoDB cursor, etc.
            unit_list = self._db.get_compute_units(unit_manager_id=self._um_id)

            for unit in unit_list:
                unit_id = str(unit["_id"])

                new_state = unit["state"]
                if unit_id in self._shared_data:
                    old_state = self._shared_data[unit_id]["data"]["state"]
                else:
                    old_state = None
                    self._shared_data_lock.acquire()
                    self._shared_data[unit_id] = {
                        'data':          unit,
                        'callbacks':     [],
                        'facade_object': None
                    }
                    self._shared_data_lock.release()

                self._shared_data_lock.acquire()
                self._shared_data[unit_id]["data"] = unit
                self._shared_data_lock.release()

                if new_state != old_state:
                    # On a state change, we fire zee callbacks.
                    logger.info("ComputeUnit '%s' state changed from '%s' to '%s'." % (unit_id, old_state, new_state))

                    # The state of the unit has changed, We call all
                    # unit-level callbacks to propagate this.
                    self.call_callbacks(unit_id, new_state)

            # After the first iteration, we are officially initialized!
            if not self._initialized.is_set():
                self._initialized.set()
                logger.debug("Worker status set to 'initialized'.")

            time.sleep(1)

        # shut down the autonomous input / output transfer worker(s)
        for worker in self._input_file_transfer_worker_pool:
            worker.terminate()
            worker.join()
            logger.debug("UnitManager.close(): %s terminated." % worker.name)

        for worker in self._output_file_transfer_worker_pool:
            worker.terminate()
            worker.join()
            logger.debug("UnitManager.close(): %s terminated." % worker.name)

    # ------------------------------------------------------------------------
    #
    def register_unit_callback(self, unit, callback_func):
        """Registers a callback function for a ComputeUnit.
        """
        unit_uid = unit.uid

        self._shared_data_lock.acquire()
        self._shared_data[unit_uid]['callbacks'].append(callback_func)
        self._shared_data_lock.release()

        # Add the facade object if missing, e.g., after a re-connect.
        if self._shared_data[unit_uid]['facade_object'] is None:
            self._shared_data_lock.acquire()
            self._shared_data[unit_uid]['facade_object'] = unit # weakref.ref(unit)
            self._shared_data_lock.release()

        # Callbacks can only be registered when the ComputeAlready has a
        # state. To partially address this shortcomming we call the callback
        # with the current ComputePilot state as soon as it is registered.
        self.call_callbacks(
            unit_uid,
            self._shared_data[unit_uid]["data"]["state"]
        )

    # ------------------------------------------------------------------------
    #
    def register_manager_callback(self, callback_func):
        """Registers a manager-level callback.
        """
        self._manager_callbacks.append(callback_func)

    # ------------------------------------------------------------------------
    #
    def get_unit_manager_data(self):
        """Returns the raw data (JSON dict) for a UnitManger.
        """
        return self._db.get_unit_manager(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_pilot_uids(self):
        """Returns the UIDs of the pilots registered with the UnitManager.
        """
        return self._db.unit_manager_list_pilots(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_uids(self):
        """Returns the UIDs of all ComputeUnits registered with the
        UnitManager.
        """
        return self._db.unit_manager_list_compute_units(self._um_id)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_states(self, unit_uids=None):
        """Returns the states of all ComputeUnits registered with the
        Unitmanager.
        """
        return self._db.get_compute_unit_states(
            self._um_id, unit_uids)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_stdout(self, compute_unit_uid):
        """Returns the stdout for a compute unit.
        """
        return self._db.get_compute_unit_stdout(compute_unit_uid)

    # ------------------------------------------------------------------------
    #
    def get_compute_unit_stderr(self, compute_unit_uid):
        """Returns the stderr for a compute unit.
        """
        return self._db.get_compute_unit_stderr(compute_unit_uid)

    # ------------------------------------------------------------------------
    #
    def add_pilots(self, pilots):
        """Links ComputePilots to the UnitManager.
        """
        # Extract the uids
        pids = []
        for pilot in pilots:
            pids.append(pilot.uid)

        self._db.unit_manager_add_pilots(unit_manager_id=self._um_id,
                                         pilot_ids=pids)

    # ------------------------------------------------------------------------
    #
    def remove_pilots(self, pilot_uids):
        """Unlinks one or more ComputePilots from the UnitManager.
        """
        self._db.unit_manager_remove_pilots(unit_manager_id=self._um_id,
                                            pilot_ids=pilot_uids)

    # ------------------------------------------------------------------------
    #
    def schedule_compute_units(self, pilot_uid, units, session):
        """Request the scheduling of one or more ComputeUnits on a
           ComputePilot.
        """

        # Get the credentials from the session.
        cred_dict = []
        for cred in session.credentials:
            cred_dict.append(cred.as_dict())

        unit_descriptions = list()
        wu_transfer = list()
        wu_notransfer = list()

        # Get some information about the pilot sandbox from the database.
        pilot_info = self._db.get_pilots(pilot_ids=pilot_uid)
        pilot_sandbox = pilot_info[0]['sandbox']

        # Split units into two different lists: the first list contains the CUs
        # that need file transfer and the second list contains the CUs that
        # don't. The latter is added to the pilot directly, while the former
        # is added to the transfer queue.
        for unit in units:
            if unit.description.input_data is None:
                wu_notransfer.append(unit.uid)
            else:
                wu_transfer.append(unit)

        # Add all units to the database.
        results = self._db.insert_compute_units(
            pilot_uid=pilot_uid,
            pilot_sandbox=pilot_sandbox,
            unit_manager_uid=self._um_id,
            units=units,
            unit_log=[]
        )

        assert len(units) == len(results)

        # Match results with units.
        for unit in units:
            # Create a shared data store entry
            self._shared_data[unit.uid] = {
                'data':          results[unit.uid],
                'callbacks':     [],
                'facade_object': unit # weakref.ref(unit)
            }

        # Bulk-add all non-transfer units-
        self._db.assign_compute_units_to_pilot(
            unit_uids=wu_notransfer,
            pilot_uid=pilot_uid
        )

        for unit in wu_notransfer:
            log = "Scheduled for execution on ComputePilot %s." % pilot_uid
            self._set_state(unit, PENDING_EXECUTION, log)

        logger.info(
            "Scheduled ComputeUnits %s for execution on ComputePilot '%s'." %
            (wu_notransfer, pilot_uid)
        )

        # Bulk-add all units that need transfer to the transfer queue.
        # Add the startup request to the request queue.
        if len(wu_transfer) > 0:
            for unit in wu_transfer:
                log = "Scheduled for data tranfer to ComputePilot %s." % pilot_uid
                self._set_state(unit.uid, PENDING_INPUT_TRANSFER, log)



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler
   :platform: Unix
   :synopsis: Scheduler implementations.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

from direct_submission import DirectSubmissionScheduler
from round_robin import RoundRobinScheduler

# -----------------------------------------------------------------------------
# Constants
SCHED_ROUND_ROBIN       = "round_robin"
SCHED_DIRECT_SUBMISSION = "direct_submission"

# -----------------------------------------------------------------------------
# 
def get_scheduler(name):
    """get_scheduler returns a scheduler object for 'name'.
    """
    if name == SCHED_ROUND_ROBIN:
        return RoundRobinScheduler()

    elif name == SCHED_DIRECT_SUBMISSION:
        return DirectSubmissionScheduler()

    else:
        raise RuntimeError("Scheduler '%s' doesn't exist." % name)



"""Pilot Manager tests
"""

import os
import sys
import radical.pilot
import unittest

import uuid
from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICALPILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICALPILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = 'radicalpilot_unittests'


#-----------------------------------------------------------------------------
#
class Test_PilotManager(unittest.TestCase):
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_create(self):
        """ Test if pilot manager creation works as expected.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        assert session.list_pilot_managers() == [], "Wrong number of pilot managers"

        pm = radical.pilot.PilotManager(session=session)
        assert session.list_pilot_managers() == [pm.uid], "Wrong list of pilot managers"

        pm = radical.pilot.PilotManager(session=session)
        assert len(session.list_pilot_managers()) == 2, "Wrong number of pilot managers"

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_reconnect(self):
        """ Test if pilot manager re-connect works as expected.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pm = radical.pilot.PilotManager(session=session)
        assert session.list_pilot_managers() == [pm.uid], "Wrong list of pilot managers"

        pm_r = session.get_pilot_managers(pilot_manager_ids=pm.uid)

        assert session.list_pilot_managers() == [pm_r.uid], "Wrong list of pilot managers"

        assert pm.uid == pm_r.uid, "Pilot Manager IDs not matching!"

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_list_pilots(self):
        """ Test if listing pilots works as expected.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pm1 = radical.pilot.PilotManager(session=session)
        assert len(pm1.list_pilots()) == 0, "Wrong number of pilots returned."

        pm2 = radical.pilot.PilotManager(session=session)
        assert len(pm2.list_pilots()) == 0, "Wrong number of pilots returned."

        for i in range(0, 2):
            cpd = radical.pilot.ComputePilotDescription()
            cpd.resource = "localhost"
            cpd.cores = 1
            cpd.runtime = 1
            cpd.sandbox = "/tmp/radical.pilot.sandbox.unittests"

            pm1.submit_pilots(pilot_descriptions=cpd)
            pm2.submit_pilots(pilot_descriptions=cpd)

        assert len(pm1.list_pilots()) == 2, "Wrong number of pilots returned."
        assert len(pm2.list_pilots()) == 2, "Wrong number of pilots returned."

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_list_pilots_after_reconnect(self):
        """ Test if listing pilots after a reconnect works as expected.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pm1 = radical.pilot.PilotManager(session=session)
        assert len(pm1.list_pilots()) == 0, "Wrong number of pilots returned."

        pm2 = radical.pilot.PilotManager(session=session)
        assert len(pm2.list_pilots()) == 0, "Wrong number of pilots returned."

        for i in range(0, 2):
            cpd = radical.pilot.ComputePilotDescription()
            cpd.resource = "localhost"
            cpd.cores = 1
            cpd.runtime = 1
            cpd.sandbox = "/tmp/radical.pilot.sandbox.unittests"

            pm1.submit_pilots(pilot_descriptions=cpd)
            pm2.submit_pilots(pilot_descriptions=cpd)

        assert len(pm1.list_pilots()) == 2, "Wrong number of pilots returned."
        assert len(pm2.list_pilots()) == 2, "Wrong number of pilots returned."

        pm1_r = session.get_pilot_managers(pilot_manager_ids=pm1.uid)
        pm2_r = session.get_pilot_managers(pilot_manager_ids=pm2.uid)

        assert len(pm1_r.list_pilots()) == 2, "Wrong number of pilots returned."
        assert len(pm2_r.list_pilots()) == 2, "Wrong number of pilots returned."

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_get_pilots(self):
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pm1 = radical.pilot.PilotManager(session=session)
        assert len(pm1.list_pilots()) == 0, "Wrong number of pilots returned."

        pm2 = radical.pilot.PilotManager(session=session)
        assert len(pm2.list_pilots()) == 0, "Wrong number of pilots returned."

        pm1_pilot_uids = []
        pm2_pilot_uids = []

        for i in range(0, 2):
            cpd = radical.pilot.ComputePilotDescription()
            cpd.resource = "localhost"
            cpd.cores = 1
            cpd.runtime = 1
            cpd.sandbox = "/tmp/radical.pilot.sandbox.unittests"

            pilot_pm1 = pm1.submit_pilots(pilot_descriptions=cpd)
            pm1_pilot_uids.append(pilot_pm1.uid)

            pilot_pm2 = pm2.submit_pilots(pilot_descriptions=cpd)
            pm2_pilot_uids.append(pilot_pm2.uid)

        for i in pm1.list_pilots():
            pilot = pm1.get_pilots(i)
            assert pilot[0].uid in pm1_pilot_uids, "Wrong pilot ID %s (not in %s)" % (pilot[0].uid, pm1_pilot_uids)


        assert len(pm1.get_pilots()) == 2, "Wrong number of pilots."

        for i in pm2.list_pilots():
            pilot = pm2.get_pilots(i)
            assert pilot[0].uid in pm2_pilot_uids, "Wrong pilot ID %s" % pilot[0].uid

        assert len(pm2.get_pilots()) == 2, "Wrong number of pilots."

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__pilotmanager_wait(self):
        """Test if wait() waits until all (2) pilots have reached 'DONE' state.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)

        pmgr = radical.pilot.PilotManager(session=session)
        
        cpd1 = radical.pilot.ComputePilotDescription()
        cpd1.resource = "localhost"
        cpd1.cores = 1
        cpd1.runtime = 1
        cpd1.sandbox = "/tmp/radical.pilot.sandbox.unittests"

        cpd2 = radical.pilot.ComputePilotDescription()
        cpd2.resource = "localhost"
        cpd2.cores = 1
        cpd2.runtime = 2
        cpd2.sandbox = "/tmp/radical.pilot.sandbox.unittests"

        pilots = pmgr.submit_pilots([cpd1, cpd2])

        pmgr.wait_pilots()
        
        for pilot in pilots:
            assert pilot.state == radical.pilot.states.DONE
            assert pilot.stop_time is not None
            assert pilot.start_time is not None

        session.close()



#pylint: disable=C0301, C0103, W0212

"""
.. module:: radical.pilot.scheduler.DirectSubmissionScheduler
   :platform: Unix
   :synopsis: A single-pilot, direct submission scheduler.

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os

from radical.pilot.utils.logger        import logger
from radical.pilot.scheduler.interface import Scheduler 

# -----------------------------------------------------------------------------
# 
class DirectSubmissionScheduler(Scheduler):
    """DirectSubmissionScheduler implements a single-pilot 'pass-through' 
    scheduling algorithm.
    """

    # -------------------------------------------------------------------------
    #
    def __init__(self):
        """Le constructeur.
        """
        Scheduler.__init__(self)
        logger.info("Loaded scheduler: %s." % self.name)

    # -------------------------------------------------------------------------
    #
    def __del__(self):
        """Le destructeur.
        """
        if os.getenv("RADICALPILOT_GCDEBUG", None) is not None:
            logger.debug("__del__(): %s." % self.name)

    # -------------------------------------------------------------------------
    #
    def _name(self):
        return "DirectSubmissionScheduler"

    # -------------------------------------------------------------------------
    #
    def schedule(self, manager, unit_descriptions):
        if manager is None:
            raise RuntimeError ('Unit scheduler is not initialized')

        pilots = manager.list_pilots()

        if not len (pilots):
            raise RuntimeError ('Unit scheduler cannot operate on empty pilot set')

        if len (pilots) > 1:
            raise RuntimeError ('Direct Submission only works for a single pilot!')
        
        ret            = dict()
        ret[pilots[0]] = list ()
        for ud in unit_descriptions:
            ret[pilots[0]].append (ud)

        return ret



""" Test resources
"""

import radical.pilot

import os
import uuid
import getpass
import unittest

from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICALPILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICALPILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = 'radicalpilot_unittests'

RESCFG = 'https://raw.github.com/radical-cybertools/radical.pilot/master/configs/futuregrid.json'

#-----------------------------------------------------------------------------
#
class TestRemoteSubmission(unittest.TestCase):
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

        self.test_resource = os.getenv('RADICALPILOT_TEST_REMOTE_RESOURCE',     "localhost")
        self.test_ssh_uid  = os.getenv('RADICALPILOT_TEST_REMOTE_SSH_USER_ID',  None)
        self.test_ssh_key  = os.getenv('RADICALPILOT_TEST_REMOTE_SSH_USER_KEY', None)
        self.test_workdir  = os.getenv('RADICALPILOT_TEST_REMOTE_WORKDIR',      "/tmp/radical.pilot.sandbox.unittests")
        self.test_cores    = os.getenv('RADICALPILOT_TEST_REMOTE_CORES',        "1")
        self.test_num_cus  = os.getenv('RADICALPILOT_TEST_REMOTE_NUM_CUS',      "2")
        self.test_timeout  = os.getenv('RADICALPILOT_TEST_TIMEOUT',             "5")

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__remote_simple_submission(self):
        """ Test simple remote submission with one pilot.
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)
        cred = radical.pilot.SSHCredential()
        cred.user_id  = self.test_ssh_uid
        cred.user_key = self.test_ssh_key

        session.add_credential(cred)

        pm = radical.pilot.PilotManager(session=session, resource_configurations=RESCFG)

        cpd = radical.pilot.ComputePilotDescription()
        cpd.resource = self.test_resource
        cpd.cores = self.test_cores
        cpd.runtime = 5
        cpd.sandbox = self.test_workdir

        pilot = pm.submit_pilots(pilot_descriptions=cpd)

        um = radical.pilot.UnitManager(session=session, scheduler='round_robin')
        um.add_pilots(pilot)

        cudescs = []
        for _ in range(0,int(self.test_num_cus)):
            cudesc = radical.pilot.ComputeUnitDescription()
            cudesc.cores = 1
            cudesc.executable = "/bin/sleep"
            cudesc.arguments = ['10']
            cudescs.append(cudesc)

        cus = um.submit_units(cudescs)

        for cu in cus:
            assert cu is not None
            assert cu.start_time is None
            assert cu.stop_time is None

        um.wait_units(state=[radical.pilot.states.DONE, radical.pilot.states.FAILED], timeout=self.test_timeout)

        for cu in cus:
            assert cu.state == radical.pilot.states.DONE
            assert cu.stop_time is not None

        pm.cancel_pilots()

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__remote_pilot_wait(self):
        """ Test if we can wait for different pilot states. 
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)
        cred = radical.pilot.SSHCredential()
        cred.user_id  = self.test_ssh_uid
        cred.user_key = self.test_ssh_key

        session.add_credential(cred)

        pm = radical.pilot.PilotManager(session=session, resource_configurations=RESCFG)

        cpd = radical.pilot.ComputePilotDescription()
        cpd.resource          = self.test_resource
        cpd.cores             = self.test_cores
        cpd.runtime           = 2
        cpd.sandbox           = self.test_workdir 

        pilot = pm.submit_pilots(pilot_descriptions=cpd)

        assert pilot is not None
        #assert cu.start_time is None
        #assert cu.start_time is None

        pilot.wait(radical.pilot.states.RUNNING, timeout=5.0*60)
        assert pilot.state == radical.pilot.states.RUNNING
        assert pilot.start_time is not None
        assert pilot.submission_time is not None


        # the pilot should finish after it has reached run_time
        pilot.wait(radical.pilot.states.DONE, timeout=5.0*60)
        assert pilot.state == radical.pilot.states.DONE
        assert pilot.stop_time is not None

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__remote_pilot_cancel(self):
        """ Test if we can cancel a pilot. 
        """
        session = radical.pilot.Session(database_url=DBURL, database_name=DBNAME)
        cred = radical.pilot.SSHCredential()
        cred.user_id  = self.test_ssh_uid
        cred.user_key = self.test_ssh_key

        session.add_credential(cred)

        pm = radical.pilot.PilotManager(session=session, resource_configurations=RESCFG)

        cpd = radical.pilot.ComputePilotDescription()
        cpd.resource          = self.test_resource
        cpd.cores             = self.test_cores
        cpd.runtime           = 2
        cpd.sandbox           = self.test_workdir 

        pilot = pm.submit_pilots(pilot_descriptions=cpd)

        assert pilot is not None
        #assert cu.start_time is None
        #assert cu.start_time is None

        pilot.wait(radical.pilot.states.RUNNING)
        assert pilot.state == radical.pilot.states.RUNNING, "Expected state 'RUNNING' but got %s" % pilot.state
        assert pilot.submission_time is not None
        assert pilot.start_time is not None

        # the pilot should finish after it has reached run_time
        pilot.cancel()

        pilot.wait(radical.pilot.states.CANCELED, timeout=5.0*60)
        assert pilot.state == radical.pilot.states.CANCELED
        assert pilot.stop_time is not None

        session.close()




"""Database conncetion layer tests
"""

import unittest

import os
import sys
import uuid
from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICALPILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICALPILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = 'radical.pilot_unittests'

#-----------------------------------------------------------------------------
#
class Test_Database():
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__new_session(self):
        """ Test if Session.new() behaves as expected.
        """
        sid = str(uuid.uuid4())
        try:
            # this should fail
            s = Session.new(sid=sid, db_url="unknownhost", db_name=DBNAME)
            assert False, "Prev. call should have failed."
        except Exception:
            assert True

        s = Session.new(sid=sid, db_url=DBURL, )
        assert s.session_id == sid
        s.delete()

    #-------------------------------------------------------------------------
    #
    def test__delete_session(self):
        """ Test if session.delete() behaves as expceted.
        """
        sid = str(uuid.uuid4())

        try:
            s = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)
            s.delete()
            assert False, "Prev. call should have failed."
        except Exception:
            assert True

    #-------------------------------------------------------------------------
    #
    def test__reconnect_session(self):
        """ Test if Session.reconnect() behaves as expceted.
        """
        sid = str(uuid.uuid4())

        try:
            s = Session.reconnect(sid=sid, db_url=DBURL, db_name=DBNAME, )
            assert False, "Prev. call should have failed."
        except Exception, ex:
            assert True

        s1 = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)
        s2 = Session.reconnect(sid=sid, db_url=DBURL, db_name=DBNAME)
        assert s1.session_id == s2.session_id
        s2.delete()

    #-------------------------------------------------------------------------
    #
    def test__insert_pilots_single(self):
        """ Tests if insert_pilots() behaves as epxected for a single pilot.
        """
        sid = str(uuid.uuid4())

        s = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)

        pilot = {
            "X": "foo",
            "Y": "bar"
        }
        ids = s.insert_pilots([pilot])
        assert len(ids) == 1, "Wrong number of IDs returned."

        pilots = s.get_raw_pilots()
        assert len(pilots) == 1, "Wrong number of pilots"
        assert pilots[0]["description"]["X"] == "foo", "Missing / wrong key."
        assert pilots[0]["state"] == "UNKNOWN", "Missing / wrong key."
        assert pilots[0]["wu_queue"] == [], "Missing / wrong key."

    #-------------------------------------------------------------------------
    #
    def test__insert_pilots_multi(self):
        """ Tests if insert_pilots() behaves as epxected for multiple pilot.
        """
        sid = str(uuid.uuid4())

        s = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)

        pilot_docs = []
        for i in range(0,32):
            pilot = {
                "X": "foo",
                "Y": str(i)
            }
            pilot_docs.append(pilot)

        ids = s.insert_pilots(pilot_docs)
        assert len(ids) == 32, "Wrong number of IDs returned."

        pilots = s.get_raw_pilots()
        for i in range(0,32):
            assert len(pilots) == 32, "Wrong number of pilots"
            assert pilots[i]["description"]["X"] == "foo", "Missing / wrong key."
            assert pilots[i]["state"] == "UNKNOWN", "Missing / wrong key."
            assert pilots[i]["wu_queue"] == [], "Missing / wrong key."

    #-------------------------------------------------------------------------
    #
    def test__insert_workunits_single(self):
        """ Tests if insert_workunits() behaves as expected for a single wu.
        """
        sid = str(uuid.uuid4())

        s = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)

        pilot = {"X": "foo",
                 "Y": "bar"
        }
        p_ids = s.insert_pilots([pilot])

        wu = {
            "description"  : {"A": "foo", "B": "bar" },
            "queue_id"     : "TODO"
        }

        wu_ids = s.insert_workunits(p_ids[0], [wu])

        workunits = s.get_raw_workunits()
        assert len(workunits) == 1, "Wrong number of workunits"
        assert workunits[0]["description"]["A"] == "foo", "Missing / wrong key."
        assert workunits[0]["assignment"]["pilot"] == p_ids[0], "Missing / wrong key."
        assert workunits[0]["state"] == "UNKNOWN", "Missing / wrong key."

        # make sure that the work units have been appended to the pilot's queue
        p = s.get_raw_pilots(p_ids)
        assert len(p[0]["wu_queue"]) == 1

    #-------------------------------------------------------------------------
    #
    def test__insert_workunits_multi(self):
        """ Tests if insert_workunits() behaves as expected for multiple wus.
        """
        sid = str(uuid.uuid4())

        s = Session.new(sid=sid, db_url=DBURL, db_name=DBNAME)

        pilot = {"X": "foo",
                 "Y": "bar"
        }
        p_ids = s.insert_pilots([pilot])

        wus = []
        for i in range(0, 128):
            wu = {
                "description"  : {"A": "foo", "B": "bar" },
                "queue_id"     : "TODO"
            }
            wus.append(wu)

        wu_ids = s.insert_workunits(p_ids[0], wus)
        assert len(wu_ids) == 128, "Wrong number of workunits"

        workunits = s.get_raw_workunits()
        assert len(workunits) == 128, "Wrong number of workunits"

        assert workunits[0]["description"]["A"] == "foo", "Missing / wrong key."
        assert workunits[0]["assignment"]["pilot"] == p_ids[0], "Missing / wrong key."
        assert workunits[0]["state"] == "UNKNOWN", "Missing / wrong key."

        # make sure that the work units have been appended to the pilot's queue
        p = s.get_raw_pilots(p_ids)
        assert len(p[0]["wu_queue"]) == 128

    #-------------------------------------------------------------------------
    #
    # def test__add_remove_work_units(self):
    #     """ Test if work_units_add(), work_units_update() and 
    #         work_units_get() behave as expected.
    #     """
    #     s = Session.new(db_url=DBURL, sid="my_new_session")

    #     wus = s.work_units_get()
    #     assert len(wus) == 0, "There shouldn't be any workunits in the collection."

    #     wu = {
    #         "work_unit_id"  : "unique work unit ID",
    #         "description"   : {
    #             "x" : "y"
    #         },
    #         "assignment"    : { 
    #             "queue" : "queue id",
    #             "pilot" : "pilot id"
    #         }
    #     }

    #     inserts = []
    #     for x in range(0,128):
    #         inserts.append(deepcopy(wu))
    #     ids = s.work_units_add(inserts)
    #     assert len(ids) == 128, "Wrong number of workunits added."




